import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import pandas as pd
import os
import math
import itertools

from torch.utils.data import Dataset, DataLoader
from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import AllChem
from rdkit import RDConfig
from rdkit.Chem import ChemicalFeatures
from rdkit.Chem.rdMolTransforms import CanonicalizeConformer
from rdkit.Chem.rdMolTransforms import GetAngleDeg, GetDihedralDeg
from rdkit.Chem.rdmolops import GetAdjacencyMatrix
from rdkit.Chem import Descriptors3D
from rdkit.Chem.rdmolfiles import CanonicalRankAtoms
from openbabel import pybel as pb
RDLogger.DisableLog('rdApp.*') # Suppress warnings from rdkit

from sac2021.const import ignored_uids, max_n_atoms, DUMMY_ATOM, a2i, hyb2int, bondtype2int

class SACData(Dataset):
    version = '0.1.0'

    def __init__(
            self,
            meta,
            data,
            idx=None,
            pretrain=False
        ):
        super(SACData, self).__init__()
        self.data_dir = data

        self.meta = pd.read_csv(meta)
        self.meta = self.meta[~self.meta.uid.isin(ignored_uids)].reset_index(drop=True)

        if idx is not None:
            self.meta = self.meta.loc[idx].reset_index(drop=True)
        self.meta = self.meta.to_records()

        # Needed for donor/acceptor features.
        fdefName = os.path.join(RDConfig.RDDataDir, 'BaseFeatures.fdef')
        self.factory = ChemicalFeatures.BuildFeatureFactory(fdefName)

        self.pretrain = pretrain

    def __getitem__(self, i):
        m = self.meta[i]

        mol = Chem.SDMolSupplier(f'{self.data_dir}/{m.uid}.sdf')[0]
        pb_mol = pb.readstring("smi", Chem.MolToSmiles(mol, isomericSmiles=False))
        conformer = mol.GetConformer()

        n_atoms = mol.GetNumAtoms()

        # Atom index features.
        atom_idx = torch.tensor(
            [self._get_atom_index(a.GetSymbol()) for a in mol.GetAtoms()] +
            [DUMMY_ATOM] + # note that a dummy atom is added after all atoms in the molecule.
            [-1] * (max_n_atoms - (n_atoms + 1)),
            dtype=torch.long
        )

        # Atom coordinate features. (max_n_atoms x 3)
        xyz = self._get_atom_xyz(mol, pad=max_n_atoms - n_atoms)

        # Mask for self-attention. 1 means masked. (1 x max_n_atoms x max_n_atoms)
        mask = torch.ones([max_n_atoms, max_n_atoms], dtype=bool)
        mask_1d = (atom_idx != -1)

        mask[mask_1d] = 0
        mask[:, mask_1d] = 0
        mask[~mask_1d] = 1
        mask[:, ~mask_1d] = 1

        mask = mask.unsqueeze(0)

        # Pairwise distance feature. (max_n_atoms x max_n_atoms)
        pdist = torch.cdist(xyz, xyz, p=2.0)
        
        # set distance to dummy node as 1e6
        pdist[n_atoms] = 1e6
        pdist[:, n_atoms] = 1e6

        # Adjacency matrix. (max_n_atoms x max_n_atoms)
        adj = torch.zeros([max_n_atoms, max_n_atoms], dtype=torch.float32)
        _adj = GetAdjacencyMatrix(mol)
        adj[:n_atoms, :n_atoms] = torch.tensor(_adj, dtype=torch.float32)

        # Angles and 2-adjacency. (max_n_atoms x max_n_atoms)
        angle = torch.zeros([max_n_atoms, max_n_atoms], dtype=torch.float32)
        adj_2 = _adj @ _adj

        for a in range(n_atoms):
            for b in range(a + 1, n_atoms):
                if adj_2[a, b] != 0:  # a, b are two-hop neighbors, so we find common neighbor and measure angles. 
                    common_neighbor = (_adj[a] * _adj[b]).argmax()
                    angle[a, b] = angle[b, a] = GetAngleDeg(conformer, a, int(common_neighbor), b) / 90.0 * math.pi

        # Torsion angles. (max_n_atoms x max_n_atoms)
        torsion_angle = torch.zeros([max_n_atoms, max_n_atoms], dtype=torch.float32)
        adj_3 = (adj_2 @ _adj > 0).astype(int) - _adj
        
        for a in range(n_atoms):
            for b in range(a + 1, n_atoms):
                if adj_3[a, b] != 0:  # a, b are three-hop neighbors.
                    a_neighbors = np.where(_adj[a] == 1)[0]
                    b_neighbors = np.where(_adj[b] == 1)[0]

                    for a_n, b_n in itertools.product(a_neighbors, b_neighbors):
                        if mol.GetBondBetweenAtoms(int(a_n), int(b_n)) is not None:
                            torsion_angle[a, b] = torsion_angle[b, a] = GetDihedralDeg(conformer, a, int(a_n), int(b_n), b) / 90.0 * math.pi
                            break
        
        # Aggregate angle features into single tensor. (max_n_atoms x max_n_atoms x 2)
        angle = torch.cat([angle.unsqueeze(2), torsion_angle.unsqueeze(2)], dim=2)

        # Bond types and conjugation.
        # Bond types (max_n_atoms x max_n_atoms x 5)
        # Conjugation (max_n_atoms x max_n_atoms)
        bond_types = torch.zeros([max_n_atoms, max_n_atoms, 5], dtype=torch.float32)
        conjugation = torch.zeros([max_n_atoms, max_n_atoms], dtype=torch.float32)
        for bond in mol.GetBonds():
            a, b = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            bt_idx = bondtype2int[bond.GetBondType()]
            bond_types[a, b, bt_idx] = bond_types[b, a, bt_idx] = 1.0
            conjugation[a, b] = conjugation[b, a] = float(bond.GetIsConjugated())

        # Symmetry class. (max_n_atoms x max_n_atoms)
        sym_groups = np.array(CanonicalRankAtoms(mol, breakTies=False))
        sym_mat = torch.zeros([max_n_atoms, max_n_atoms], dtype=torch.bool)
        for i, group in enumerate(sym_groups):
            sym_mat[i, [idx for idx in np.where(sym_groups == group)[0] if idx != i]] = 1

        # One-hot atom hybridization (7 types) (max_n_atoms x 7)
        hyb = self._get_atom_hybridization(mol)
        # Is an atom contained in aromatic ring? (max_n_atoms x 1)
        is_aromatic = self._get_atom_is_aromatic(mol)
        # Formal charge. (max_n_atoms x 1)
        formal_charge = self._get_formal_charge(mol)
        # Total_numh. (max_n_atoms x 1)
        total_numh = self._get_total_num_H(mol)
        # Total_valence. (max_n_atoms x 1)
        total_valence = self._get_total_valence(mol)
        # One-hot label for donor & acceptor status. (max_n_atoms x 2)
        donor_acceptor = self._get_donor_acceptor(mol)

        spin = torch.zeros([max_n_atoms, 2])
        spin[:, pb_mol.spin - 1] = 1 # Spin multiplicity

        # Molecule-level features.
        npr1 = torch.ones([max_n_atoms, 1]) * Descriptors3D.NPR1(mol) # NPR1
        npr2 = torch.ones([max_n_atoms, 1]) * Descriptors3D.NPR2(mol) # NPR2

        fp2 = pb_mol.calcfp(fptype='FP2')
        fp = torch.zeros([1021], dtype=torch.float)
        for b in fp2.bits:
            fp[b - 1] = 1.0

        feat = torch.cat([
            is_aromatic,    # dim=1
            formal_charge,  # dim=1
            total_numh,     # dim=1
            total_valence,  # dim=1
            npr1,           # dim=1
            npr2,           # dim=1
        ], dim=-1)          # total dim=6

        item = {
            'mask': mask.bool(),
            'out_mask': mask_1d.bool(),
            'n_atoms': mask_1d.sum(),
            'atom_idx': atom_idx.long(),
            'hyb': hyb,
            'donac': donor_acceptor,
            'spin': spin,
            'pdist': pdist,
            'sym': sym_mat,
            'angle': angle,
            'adj': adj,
            'feat': feat,
            'fp': fp,
        }

        for k, v in item.items():
            if torch.isnan(v).any():
                print(k, 'has NaN value!')
                print('INFO:')
                print(m)

        if self.pretrain:
            item.update({
                'homo': m['homo'],
                'lumo': m['lumo'],
                'target': m['lumo'] - m['homo'],
            })
        else:
            item.update({
                's1': m['S1_energy(eV)'],
                't1': m['T1_energy(eV)'],
                'target': m['S1_energy(eV)'] - m['T1_energy(eV)'],
            })

        return item

    def _get_donor_acceptor(self, mol):
        v = torch.zeros([max_n_atoms, 2])
        feats = self.factory.GetFeaturesForMol(mol)
        for feat in feats:
            if feat.GetFamily() == 'Donor':
                for a_idx in feat.GetAtomIds():
                    v[a_idx, 0] = 1

            if feat.GetFamily() == 'Acceptor':
                for a_idx in feat.GetAtomIds():
                    v[a_idx, 1] = 1
        
        return v

    def _get_total_valence(self, mol):
        # valence = 0, 1, 2, 3, 4, 5, 6
        v = torch.zeros([max_n_atoms, 1])
        for i, atom in enumerate(mol.GetAtoms()):
            v[i] = atom.GetTotalValence() / 1.5 - 2.
        return v

    def _get_total_num_H(self, mol):
        # total_num_h = 0, 1, 2, 3, 4
        v = torch.zeros([max_n_atoms, 1])
        for i, atom in enumerate(mol.GetAtoms()):
            v[i] = atom.GetTotalNumHs(includeNeighbors=True) - 2.
        return v

    def _get_formal_charge(self, mol):
        # formal_charge = -3, -2, -1, 0, 1, 2, 3
        v = torch.zeros([max_n_atoms, 1])
        for i, atom in enumerate(mol.GetAtoms()):
            v[i] = atom.GetFormalCharge() / 1.5
        return v

    def _get_atom_is_aromatic(self, mol):
        onehot = torch.zeros([max_n_atoms, 1])
        for i, atom in enumerate(mol.GetAtoms()):
            if atom.GetIsAromatic():
                onehot[i, 0] = 1.

        return onehot

    def _get_atom_hybridization(self, mol):
        hyb_onehot = torch.zeros([max_n_atoms, 7])
        for i, atom in enumerate(mol.GetAtoms()):
            hyb_onehot[i, hyb2int.get(atom.GetHybridization(), 6)] = 1.

        return hyb_onehot
    
    def _get_atom_xyz(self, mol, pad):
        conf = mol.GetConformer()
        CanonicalizeConformer(conf)
        position = torch.tensor(conf.GetPositions()).float()
        return torch.cat([position, torch.zeros([pad, 3])], dim=0)

    def _get_atom_index(self, a):
        return a2i.get(a, 2)
    
    def __len__(self):
        return len(self.meta)


class SACDataInfer(SACData):
    version = '0.1.0'

    def __init__(
            self,
            meta,
            data,
            idx=None,
            pretrain=False
        ):
        super(SACData, self).__init__()
        self.meta = pd.read_csv(meta)
        self.meta = self.meta[~self.meta.uid.isin(ignored_uids)].reset_index(drop=True)
        if idx is not None:
            self.meta = self.meta.loc[idx].reset_index(drop=True)
        self.meta = self.meta.to_records()
        
        self.data_dir = data

        fdefName = os.path.join(RDConfig.RDDataDir, 'BaseFeatures.fdef')
        self.factory = ChemicalFeatures.BuildFeatureFactory(fdefName) # Needed for donor/acceptor features.

        self.pretrain = pretrain

    def __getitem__(self, i):
        m = self.meta[i]
        mol = Chem.SDMolSupplier(f'{self.data_dir}/{m.uid}.sdf')[0]
        pb_mol = pb.readstring("smi", Chem.MolToSmiles(mol))

        n_atoms = mol.GetNumAtoms()

        # Atom index features.
        atom_idx = torch.tensor(
            [self._get_atom_index(a.GetSymbol()) for a in mol.GetAtoms()] +
            [DUMMY_ATOM] + # note that a dummy atom is added after all atoms in the molecule.
            [-1] * (max_n_atoms - n_atoms - 1),
            dtype=torch.long
        )

        # Atom coordinate features.
        xyz = self._get_atom_xyz(mol, pad=max_n_atoms - n_atoms)

        mask = torch.ones([max_n_atoms, max_n_atoms], dtype=bool)
        mask_1d = (atom_idx != -1)
        mask[mask_1d] = 0
        mask[:, mask_1d] = 0
        mask[~mask_1d] = 1
        mask[:, ~mask_1d] = 1
        mask = mask.unsqueeze(0)

        # pairwise distance
        pdist = torch.cdist(xyz, xyz, p=2.0)
        # set distance to dummy node as 1e6
        pdist[n_atoms] = 1e6
        pdist[:, n_atoms] = 1e6

        # adjacency
        adj = torch.zeros([max_n_atoms, max_n_atoms], dtype=bool)
        adj[:n_atoms, :n_atoms] = torch.tensor(GetAdjacencyMatrix(mol), dtype=bool)

        # atom hybridization (one-hot, 7 types) : max_n_atoms x 7
        hyb = self._get_atom_hybridization(mol)
        # is_aromatic : max_n_atoms x 1
        is_aromatic = self._get_atom_is_aromatic(mol)
        # formal_charge : max_n_atoms x 1
        formal_charge = self._get_formal_charge(mol)
        # total_numh : max_n_atoms x 1
        total_numh = self._get_total_num_H(mol)
        # total_valence : max_n_atoms x 1
        total_valence = self._get_total_valence(mol)
        # donor & acceptor
        donor_acceptor = self._get_donor_acceptor(mol)

        # Molecule-level features.
        spin = torch.zeros([max_n_atoms, 2])
        spin[:, pb_mol.spin - 1] = 1 # Spin multiplicity

        npr1 = torch.ones([max_n_atoms, 1]) * Descriptors3D.NPR1(mol) # NPR1
        npr2 = torch.ones([max_n_atoms, 1]) * Descriptors3D.NPR2(mol) # NPR2

        feat = torch.cat([
            hyb,
            is_aromatic,
            formal_charge,
            total_numh,
            total_valence,
            donor_acceptor,
            spin,
            npr1,
            npr2
        ], dim=-1)

        item = {
            'mask': mask.bool(),
            'out_mask': mask_1d.bool(),
            'n_atoms': mask_1d.sum(),
            'atom_idx': atom_idx.long(),
            'pdist': pdist,
            'adj': adj,
            'feat': feat,
        }
        return item

if __name__ == '__main__':
    from tqdm import tqdm

    dataset = SACData(
        meta='data/pretrain_metadata.csv',
        data='data/pretrain_sdf',
        pretrain=True
    )
    loader = DataLoader(dataset, batch_size=16, num_workers=16)

    for data in tqdm(loader):
        pass

    print(f'len={len(dataset)}')
    print(SACData.version)



import torch
import torch.nn as nn

# https://github.com/tuantle/regression-losses-pytorch/blob/master/regression_losses.py
class LogCoshLoss(nn.Module):
    def __init__(self):
        super(LogCoshLoss, self).__init__()

    def forward(self, y_t, y_prime_t):
        ey_t = y_t - y_prime_t
        return torch.mean(torch.log(torch.cosh(ey_t + 1e-12)))


from rdkit import Chem
from collections import defaultdict

max_n_atoms = 148
# max_n_atoms = 310
num_unique_atoms = DUMMY_ATOM = 12
atoms = ['H', 'B', 'C', 'N', 'O', 'F', 'S', 'Si', 'P', 'Cl', 'Br', 'I']
a2i = {s:i for i, s in enumerate(atoms)}

fp = {
    'train_meta': '../../data/traindev.csv',
    'test_meta': '../../data/test_no590.csv',

    'train_data': '../../data/pretrain_sdf',
    'test_data': '../../data/test_sdf',

    'pretrain_meta': '../../data/qm9+oe62.csv',
    'pretrain_data': '../../data/pretrain_sdf',

    'sample_submission': '../../data/sample_submission_no590.csv',
}

ignored_uids = [
    'train_1688',
    'train_14782',
    'train_28906',
    'train_29068',

    'train_29036',  # cannot apply CanonicalizeConformer
    'train_29828',  # cannot apply CanonicalizeConformer

    'gdb_1',
    'gdb_2',
    'gdb_3',
    
    # Errors in angle computation.
    'gdb_23',
    'gdb_24',
    'gdb_486',
    'PAMMAS',
    'QUFCID',
    'LOVWOI',
    'QEYQUH',
    'ENIGEO',
]

hyb2int = {
    Chem.rdchem.HybridizationType.S: 0,
    Chem.rdchem.HybridizationType.SP: 1,
    Chem.rdchem.HybridizationType.SP2: 2,
    Chem.rdchem.HybridizationType.SP3: 3,
    Chem.rdchem.HybridizationType.SP3D: 4,
    Chem.rdchem.HybridizationType.SP3D2: 5,
    Chem.rdchem.HybridizationType.UNSPECIFIED: 6,
}

bondtype2int = defaultdict(lambda: 4)
bondtype2int[Chem.rdchem.BondType.SINGLE] = 0
bondtype2int[Chem.rdchem.BondType.AROMATIC] = 1
bondtype2int[Chem.rdchem.BondType.DOUBLE] = 2
bondtype2int[Chem.rdchem.BondType.TRIPLE] = 3



import wandb

import argparse
import random
import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import pandas as pd

from torch.utils.data import DataLoader
from sac2021.sac2021 import AtomTransformerPretrain as AtomTransformer
from sac2021.data import SACData
from sklearn.model_selection import KFold

import sac2021.const as const
import sac2021.loss as loss

from sac2021.scheduler import GradualWarmupScheduler

torch.autograd.set_detect_anomaly(True)

def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--meta', help='Path to metadata csv file.', required=True)
    parser.add_argument('--data', help='Path to directory containing sdf files.', required=True)
    parser.add_argument('--output', '-o', required=True)
    parser.add_argument('--model-id', required=True)
    parser.add_argument('--fold', type=int, required=True)
    parser.add_argument('--loss', required=True)
    parser.add_argument('--epochs', type=int, default=52)
    parser.add_argument('--bsz', type=int, default=64)
    parser.add_argument('--n-layers', type=int, default=24)
    parser.add_argument('--d-model', type=int, default=256)
    parser.add_argument('--n-heads', type=int, default=16)
    parser.add_argument('--d-ff', type=int, default=2048)
    parser.add_argument('--lr', type=float, default=3e-5)
    parser.add_argument('--seed', type=int, default=42)
    parser.add_argument('--use-wandb', action='store_true', default=False)
    parser.add_argument('--debug', action='store_true', default=False)

    return parser.parse_args()

args = parse_arguments()

random.seed(args.seed)
np.random.seed(args.seed)
os.environ["PYTHONHASHSEED"] = str(args.seed)
torch.manual_seed(args.seed)
torch.cuda.manual_seed(args.seed)

if not args.use_wandb:
    os.environ['WANDB_MODE'] = 'disabled'

if args.debug:
    os.environ['WANDB_MODE'] = 'offline'

meta = pd.read_csv(args.meta)
meta = meta[~meta.uid.isin(const.ignored_uids)].reset_index(drop=True)

losses = {
    'mse': nn.MSELoss,
    'mae': nn.L1Loss,
    'logcosh': loss.LogCoshLoss,
}
criterion = losses[args.loss]()
mae = nn.L1Loss()

# W&B configuration.
wandb.init(project='sac-solution', entity='dohlee', tags=['pretrain'])
config = wandb.config
config.update(args)
config.update({'pretrained': False})
config.update({'data_version': SACData.version})

# Hold out 2.5% of the dataset for validation.
cv = KFold(n_splits=40, shuffle=True, random_state=args.seed)
for fold, (train_idx, val_idx) in enumerate(cv.split(meta), 1):
    if fold == args.fold:
        break

val_idx_set = set(val_idx)
train_idx = [idx for idx in range(len(meta)) if idx not in val_idx_set]

print('train', len(train_idx))
print('val', len(val_idx))

net = AtomTransformer(
    config.n_layers,
    config.n_heads,
    config.d_model,
    config.d_ff
)
net.cuda()

wandb.watch(net, log='all')
optimizer = torch.optim.AdamW(net.parameters(), lr=args.lr)
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
    optimizer, mode='min', factor=0.5, 
    patience=10, threshold=0.005, threshold_mode='abs'
)

train_dataset = SACData(
    meta=args.meta,
    data=args.data,
    idx=train_idx,
    pretrain=True
)
val_dataset = SACData(
    meta=args.meta,
    data=args.data,
    idx=val_idx,
    pretrain=True
)
train_loader = DataLoader(
    train_dataset, num_workers=16, batch_size=config.bsz, shuffle=True, pin_memory=True,
)
val_loader = DataLoader(
    val_dataset, num_workers=16, batch_size=config.bsz, shuffle=False, pin_memory=True, drop_last=False,
)

optimizer.zero_grad()
optimizer.step()

best_val_mae, val_mae = 10000, 10000
for epoch in range(1, args.epochs + 1):
    scheduler.step(val_mae)

    # losses
    running_homo_loss = 0.0
    running_lumo_loss = 0.0

    # metrics (MAE)
    running_homo_mae = 0.0
    running_lumo_mae = 0.0

    net.train()
    for batch, data in enumerate(train_loader, 1):
        for k, v in data.items():
            data[k] = v.cuda()

        optimizer.zero_grad()

        atom_idx = data['atom_idx']
        hyb = data['hyb']
        donac = data['donac']
        spin = data['spin']
        feat = data['feat']
        pdist = data['pdist']
        angle = data['angle']
        adj = data['adj']
        mask = data['mask']
        out_mask = data['out_mask']
        n_atoms = data['n_atoms']
        homo_target = data['homo'].float()
        lumo_target = data['lumo'].float()

        homo_out, lumo_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)
        _bsz = homo_out.size(0)
        homo_loss = criterion(homo_out, homo_target.view(_bsz, 1))
        lumo_loss = criterion(lumo_out, lumo_target.view(_bsz, 1))

        loss = 0.5 * (homo_loss + lumo_loss)

        running_homo_loss += homo_loss.detach().item()
        running_lumo_loss += lumo_loss.detach().item()

        _mae = mae(homo_out, homo_target.view(_bsz, 1))
        running_homo_mae += _mae.detach().item()

        _mae = mae(lumo_out, lumo_target.view(_bsz, 1))
        running_lumo_mae += _mae.detach().item()

        loss.backward()
        torch.nn.utils.clip_grad_norm_(net.parameters(), 5) # gradient clipping.
        optimizer.step()

        if batch % 10 == 0:
            batch_homo_loss = running_homo_loss / 10.
            batch_lumo_loss = running_lumo_loss / 10.

            batch_homo_mae = running_homo_mae / 10.
            batch_lumo_mae = running_lumo_mae / 10.
            batch_homo_lmae = np.log(batch_homo_mae)
            batch_lumo_lmae = np.log(batch_lumo_mae)

            print(f'E{epoch:<3}B{batch:<3} hl={batch_homo_loss:.4f} ll={batch_lumo_loss:.4f} hmae={batch_homo_mae:.4f} lmae={batch_lumo_mae:.4f} lhmae={batch_homo_lmae:.4f} llmae={batch_lumo_lmae:.4f}')

            running_homo_loss = 0.0
            running_lumo_loss = 0.0
            running_homo_mae = 0.0
            running_lumo_mae = 0.0
            
            log_dict = {
                'train/homo_loss': batch_homo_loss,
                'train/lumo_loss': batch_lumo_loss,
                'train/homo_mae': batch_homo_mae, 
                'train/lumo_mae': batch_lumo_mae,
                'train/lr': get_lr(optimizer)
            }
            wandb.log(log_dict)

    net.eval()
    val_homo_outs, val_lumo_outs = [], []
    val_homo_targets, val_lumo_targets = [], []
    with torch.no_grad():
        for batch, data in enumerate(val_loader, 1):
            for k, v in data.items():
                data[k] = v.cuda()

            atom_idx = data['atom_idx']
            hyb = data['hyb']
            donac = data['donac']
            spin = data['spin']
            feat = data['feat']
            pdist = data['pdist']
            angle = data['angle']
            adj = data['adj']
            mask = data['mask']
            out_mask = data['out_mask']
            n_atoms = data['n_atoms']
            homo = data['homo'].float()
            lumo = data['lumo'].float()

            homo_out, lumo_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)

            _bsz = homo_out.size(0)
            val_homo_outs.append(homo_out.cpu())
            val_lumo_outs.append(lumo_out.cpu())
            val_homo_targets.append(homo.cpu().view(_bsz, -1))
            val_lumo_targets.append(lumo.cpu().view(_bsz, -1))

    val_homo_outs, val_lumo_outs = map(lambda x: torch.cat(x, dim=0), (val_homo_outs, val_lumo_outs))
    val_homo_targets, val_lumo_targets = map(lambda x: torch.cat(x, dim=0), (val_homo_targets, val_lumo_targets))

    homo_loss = criterion(val_homo_outs, val_homo_targets)
    lumo_loss = criterion(val_lumo_outs, val_lumo_targets)

    homo_mae = mae(val_homo_outs, val_homo_targets)
    lumo_mae = mae(val_lumo_outs, val_lumo_targets)
    homo_lmae = torch.log(homo_mae)
    lumo_lmae = torch.log(lumo_mae)

    print(f'***Validation E{epoch:<3} hl={homo_loss:.4f} ll={lumo_loss:.4f} hmae={homo_mae:.4f} lmae={lumo_mae:4f} hlmae={homo_lmae:.4f} llmae={lumo_lmae:.4f}')
    log_dict = {
        'val/homo_loss': homo_loss,
        'val/lumo_loss': lumo_loss,
        'val/homo_mae': homo_mae,
        'val/lumo_mae': lumo_mae,
        'epoch': epoch,
    }
    wandb.log(log_dict)

    if val_mae < best_val_mae:
        best_val_mae = val_mae
        wandb.summary['best_val_mae'] = best_val_mae

        ckpt = {
            'net': net.state_dict(),
            'optimizer': optimizer.state_dict(),
            'epoch': epoch,
            'val_mae': best_val_mae,
        }
        torch.save(ckpt, args.output)

    if epoch % 3 == 0:
        ckpt = {
            'net': net.state_dict(),
            'optimizer': optimizer.state_dict(),
            'epoch': epoch,
            'val_mae': best_val_mae,
        }
        torch.save(ckpt, args.output +f'.e{epoch}')




import wandb

import argparse
import random
import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import pandas as pd

from torch.utils.data import DataLoader
from sac2021.sac2021 import AtomTransformer
from sac2021.data import SACData
from sklearn.model_selection import KFold

import sac2021.const as const
import sac2021.loss as loss
from sac2021.scheduler import GradualWarmupScheduler

torch.autograd.set_detect_anomaly(True)

def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--output', '-o', required=True)
    # parser.add_argument('--ckpt', required=True)
    parser.add_argument('--model-id', required=True)
    parser.add_argument('--fold', type=int, required=True)
    parser.add_argument('--loss', required=True)
    parser.add_argument('--bsz', type=int, default=64)
    parser.add_argument('--n-layers', type=int, default=24)
    parser.add_argument('--d-model', type=int, default=256)
    parser.add_argument('--n-heads', type=int, default=16)
    parser.add_argument('--d-ff', type=int, default=2048)
    parser.add_argument('--lr', type=float, default=3e-5)
    parser.add_argument('--res-p', type=float, default=0.0)
    parser.add_argument('--att-p', type=float, default=0.0)
    parser.add_argument('--seed', type=int, default=42)
    parser.add_argument('--augs', nargs='+', default=[])
    parser.add_argument('--debug', action='store_true', default=False)

    return parser.parse_args()

args = parse_arguments()

random.seed(args.seed)
np.random.seed(args.seed)
os.environ["PYTHONHASHSEED"] = str(args.seed)
torch.manual_seed(args.seed)
torch.cuda.manual_seed(args.seed)  # type: ignore
# torch.backends.cudnn.deterministic = True  # type: ignore
# torch.backends.cudnn.benchmark = True  # type: ignore

if args.debug:
    os.environ['WANDB_MODE'] = 'offline'

meta = pd.read_csv(const.fp['train_meta'])
meta = meta[~meta.uid.isin(const.ignored_uids)].reset_index(drop=True)

td_meta = meta[(meta.uid.str.startswith('train')) | (meta.uid.str.startswith('dev'))].reset_index(drop=True)

losses = {
    'mse': nn.MSELoss,
    'mae': nn.L1Loss,
    'logcosh': loss.LogCoshLoss,
}
criterion = losses[args.loss]()
mae = nn.L1Loss()

# W&B configuration.
wandb.init(project='sac', entity='dohlee')
config = wandb.config
config.update(args)
config.update({'pretrained': False})
# config.update({'pretrained_model': args.ckpt})
config.update({'data_version': SACData.version})

cv = KFold(n_splits=40, shuffle=True, random_state=args.seed)

for fold, (train_idx, val_idx) in enumerate(cv.split(td_meta), 1):
    if fold == args.fold:
        break

val_idx_set = set(val_idx)
train_idx = [idx for idx in range(len(meta)) if idx not in val_idx_set]

print('train', len(train_idx))
print('val', len(val_idx))

net = AtomTransformer(config.n_layers, config.n_heads, config.d_model, config.d_ff, res_p=args.res_p, att_p=args.att_p)
# Load checkpoint.
# ckpt = torch.load(args.ckpt)
# net.load_state_dict(ckpt['net'], strict=False)
# Freeze parameters. Will be unfreezed later.
# for param in net.atom_embedding.parameters():
    # param.requires_grad = False
# for param in net.transformer.parameters():
    # param.requires_grad = False

# for i in [-1, -2, -3]:
    # for param in net.transformer.layers[i].parameters():
        # param.requires_grad = True
net.cuda()

wandb.watch(net, log='all')
optimizer = torch.optim.AdamW(net.parameters(), lr=args.lr)
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
    optimizer, mode='min', factor=0.5, 
    patience=15, threshold=0.005, threshold_mode='rel'
)

train_dataset = SACData(idx=train_idx, augs=args.augs)
val_dataset = SACData(idx=val_idx)
train_loader = DataLoader(
    train_dataset, num_workers=16, batch_size=config.bsz, shuffle=True, pin_memory=True, drop_last=True,
)
val_loader = DataLoader(
    val_dataset, num_workers=16, batch_size=config.bsz, shuffle=False, pin_memory=True, drop_last=False,
)
n_batch = len(train_loader)

optimizer.zero_grad()
optimizer.step()

best_val_mae, val_mae = 10000, 10000
for epoch in range(1, 501):
    scheduler.step(val_mae)
    # scheduler.step(epoch=epoch, metrics=val_mae)
    # print(f'LR={get_lr(optimizer)}')
    # print(scheduler.get_lr())

    # losses
    running_gap_loss = 0.0
    running_s1_loss = 0.0
    running_t1_loss = 0.0

    # metrics (MAE)
    running_gap_mae = 0.0
    running_gap_mae_secondary = 0.0
    running_gap_mae_tertiary = 0.0

    net.train()
    for batch, data in enumerate(train_loader, 1):
        for k, v in data.items():
            data[k] = v.cuda()

        optimizer.zero_grad()

        atom_idx = data['atom_idx']
        hyb = data['hyb']
        donac = data['donac']
        spin = data['spin']
        feat = data['feat']
        pdist = data['pdist']
        sym = data['sym']
        angle = data['angle']
        adj = data['adj']
        mask = data['mask']
        fp = data['fp']
        out_mask = data['out_mask']
        n_atoms = data['n_atoms']
        gap_target = data['target'].float()
        s1_target = data['s1'].float()
        t1_target = data['t1'].float()

        gap_out, s1_out, t1_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)
        _bsz = gap_out.size(0)
        gap_loss = criterion(gap_out, gap_target.view(_bsz, 1))
        s1_loss = criterion(s1_out, s1_target.view(_bsz, 1))
        t1_loss = criterion(t1_out, t1_target.view(_bsz, 1))
        loss = gap_loss + 0.05 * (s1_loss + t1_loss)
        # loss = s1_loss + t1_loss

        running_gap_loss += gap_loss.detach().item()
        running_s1_loss += s1_loss.detach().item()
        running_t1_loss += t1_loss.detach().item()

        _mae = mae(gap_out, gap_target.view(_bsz, 1))
        running_gap_mae += _mae.detach().item()

        _mae = mae(s1_out - t1_out, gap_target.view(_bsz, 1))
        running_gap_mae_secondary += _mae.detach().item()

        _mae = mae(0.5 * gap_out + 0.5 * (s1_out - t1_out) , gap_target.view(_bsz, 1))
        running_gap_mae_tertiary += _mae.detach().item()

        loss.backward()
        nn.utils.clip_grad_norm_(net.parameters(), 1)
        optimizer.step()

        if batch % 10 == 0:
            batch_gap_loss = running_gap_loss / 10.
            batch_s1_loss = running_s1_loss / 10.
            batch_t1_loss = running_t1_loss / 10.
            batch_gap_mae = running_gap_mae / 10.
            batch_gap_mae_secondary = running_gap_mae_secondary / 10.
            batch_gap_mae_tertiary = running_gap_mae_tertiary / 10.

            print(f'E{epoch:<3}B{batch:<3} (lr={get_lr(optimizer):.3g})l={batch_gap_loss:.4f} l2={batch_s1_loss:.4f} l3={batch_t1_loss:.4f} mae={batch_gap_mae:.4f} mae2={batch_gap_mae_secondary:.4f} mae3={batch_gap_mae_tertiary:.4f} lmae={np.log(batch_gap_mae):.4f}')

            running_gap_loss = 0.0
            running_s1_loss = 0.0
            running_t1_loss = 0.0
            running_gap_mae = 0.0
            running_gap_mae_secondary = 0.0
            running_gap_mae_tertiary = 0.0
            
            log_dict = {
                'train/loss': batch_gap_loss,
                'train/s1_loss': batch_s1_loss,
                'train/t1_loss': batch_t1_loss,
                'train/mae': batch_gap_mae, 
                'train/mae_secondary': batch_gap_mae_secondary,
                'train/mae_tertiary': batch_gap_mae_tertiary,
                'train/lr': get_lr(optimizer)
            }
            wandb.log(log_dict)

    net.eval()
    val_gap_outs, val_s1_outs, val_t1_outs = [], [], []
    val_gap_targets, val_s1_targets, val_t1_targets = [], [], []
    with torch.no_grad():
        for batch, data in enumerate(val_loader, 1):
            for k, v in data.items():
                data[k] = v.cuda()

            atom_idx = data['atom_idx']
            hyb = data['hyb']
            donac = data['donac']
            spin = data['spin']
            feat = data['feat']
            pdist = data['pdist']
            sym = data['sym']
            angle = data['angle']
            adj = data['adj']
            mask = data['mask']
            fp = data['fp']
            out_mask = data['out_mask']
            n_atoms = data['n_atoms']
            target = data['target'].float()
            s1 = data['s1'].float()
            t1 = data['t1'].float()

            gap_out, s1_out, t1_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)

            _bsz = gap_out.size(0)
            val_gap_outs.append(gap_out.cpu())
            val_s1_outs.append(s1_out.cpu())    
            val_t1_outs.append(t1_out.cpu())
            val_gap_targets.append(target.cpu().view(_bsz, -1))
            val_s1_targets.append(s1.cpu().view(_bsz, -1))
            val_t1_targets.append(t1.cpu().view(_bsz, -1))

    val_gap_outs, val_s1_outs, val_t1_outs = map(lambda x: torch.cat(x, dim=0), (val_gap_outs, val_s1_outs, val_t1_outs))
    val_gap_targets, val_s1_targets, val_t1_targets = map(lambda x: torch.cat(x, dim=0), (val_gap_targets, val_s1_targets, val_t1_targets))

    val_gap_loss = criterion(val_gap_outs, val_gap_targets)
    val_s1_loss = criterion(val_s1_outs, val_s1_targets)
    val_t1_loss = criterion(val_t1_outs, val_t1_targets)

    val_mae = mae(val_gap_outs, val_gap_targets)
    val_mae_secondary = mae(val_s1_outs - val_t1_outs, val_gap_targets)
    val_mae_tertiary = mae(0.5 * val_gap_outs + 0.5 * (val_s1_outs - val_t1_outs), val_gap_targets)

    print(f'***Validation E{epoch:<3} loss={val_gap_loss:.4f} s1_loss={val_s1_loss:.4f} t1_loss={val_t1_loss:.4f} mae={val_mae:.4f} mae2={val_mae_secondary:.4f} mae3={val_mae_tertiary:.4f} lmae={np.log(val_mae):.4f}')
    log_dict = {
        'val/loss': val_gap_loss,
        'val/mae': val_mae,
        'val/mae_secondary': val_mae_secondary,
        'val/mae_tertiary': val_mae_tertiary,
        'epoch': epoch,
    }
    wandb.log(log_dict)

    if val_mae < best_val_mae:
        best_val_mae = val_mae
        wandb.summary['best_val_mae'] = best_val_mae

        ckpt = {
            'net': net.state_dict(),
            'optimizer': optimizer.state_dict(),
            'epoch': epoch,
            'val_mae': best_val_mae,
        }
        torch.save(ckpt, args.output)



import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from sac2021.const import max_n_atoms, num_unique_atoms

class MultiHeadAttention(nn.Module):
    def __init__(self, n_heads, d_model, idx, res_p=0.0, att_p=0.0, bias=True, scale=True):
        super(MultiHeadAttention, self).__init__()

        d_head = d_model // n_heads
        self.n_heads, self.d_head = n_heads, d_head
        self.bias, self.scale = bias, scale
        
        self.gamma_p = nn.Parameter(torch.ones([n_heads]))
        self.gamma_adj = nn.Parameter(torch.ones([n_heads]))

        self.w_bias = nn.Linear(2, n_heads, bias=False)

        self.att = nn.Linear(d_model, 3 * n_heads * d_head, bias=False)  # LinearNoBias for attention
        self.ff = nn.Linear(n_heads * d_head, d_model, bias=bias)

        self.drop_att, self.drop_res = nn.Dropout(att_p), nn.Dropout(res_p)
        self.ln = nn.LayerNorm(d_model)

        self.idx = idx

    def forward(self, x, pdist, angle, adj, mask=None):
    # def forward(self, x, pdist, sym, angle, adj, mask=None):
        ff_out = self.ff(self._attention(x, pdist, angle, adj, mask=mask))
        # ff_out = self.ff(self._attention(x, pdist, sym, angle, adj, mask=mask))
        return self.ln(x + self.drop_res(ff_out))

    # def _attention(self, x, pdist, sym, angle, adj, mask):
    def _attention(self, x, pdist, angle, adj, mask):
        # x : bsz x n_atoms x d_model
        bsz, n_atoms = x.size(0), x.size(1)
        # self.att(x) : bsz x n_atoms x (3 * n_heads * d_head)
        wq, wk, wv = torch.chunk(self.att(x), 3, dim=-1) 
        # --> wq, wk, wv : bsz x n_atoms x (n_heads * d_head)
        wq, wk, wv = map(lambda x: x.view(bsz, n_atoms, self.n_heads, self.d_head), (wq, wk, wv))
        # --> wq, wk, wv : bsz x n_atoms x n_heads x d_head
        wq, wk, wv = wq.permute(0, 2, 1, 3), wk.permute(0, 2, 3, 1), wv.permute(0, 2, 1, 3)
        # --> wq : bsz x n_heads x n_atoms x d_head
        # --> wk : bsz x n_heads x d_head x n_atoms
        # --> wv : bsz x n_heads x n_atoms x d_head
        att_score = torch.matmul(wq, wk)
        # --> att_score : bsz x n_heads x n_atoms x n_atoms
        if self.scale:
            att_score.div_(self.d_head ** 0.5)

        # Penalize by pairwise distance
        # pdist : bsz x 2 x n_atoms x n_atoms
        # --> expand to bsz x n_heads x n_atoms x n_atoms
        # --> also expand gamma to have shape bsz x n_heads x n_atoms x n_atoms

        pdist = pdist.unsqueeze(1).expand(bsz, self.n_heads, n_atoms, n_atoms)
        gamma_p = self.gamma_p.unsqueeze(1).unsqueeze(2).expand(self.n_heads, n_atoms, n_atoms).unsqueeze(0).expand(bsz, self.n_heads, n_atoms, n_atoms)
        att_score -= gamma_p * pdist

        angle = angle.view(-1, 2) # (bsz * n_atoms * n_atoms) x 2
        angle = self.w_bias(angle) # (bsz * n_atoms * n_atoms) x n_heads
        angle = angle.view(bsz, n_atoms, n_atoms, self.n_heads).permute(0, 3, 1, 2)
        att_score += angle

        adj = adj.unsqueeze(1).expand(bsz, self.n_heads, n_atoms, n_atoms)
        gamma_adj = self.gamma_adj.unsqueeze(1).unsqueeze(2).expand(self.n_heads, n_atoms, n_atoms).unsqueeze(0).expand(bsz, self.n_heads, n_atoms, n_atoms)
        att_score += gamma_adj * adj

        if mask is not None:
            minus_inf = -65504 if att_score.dtype == torch.float16 else -1e9
            att_score = att_score.masked_fill(mask, minus_inf).type_as(att_score)
        att_prob = self.drop_att(F.softmax(att_score, dim=-1))
        # --> att_prob : bsz x n_heads x n_atoms x n_atoms

        att_vec = torch.matmul(att_prob, wv)
        # --> att_vec : bsz x n_heads x n_atoms x d_head
        att_vec = att_vec.permute(0, 2, 1, 3).contiguous().contiguous().view(bsz, n_atoms, -1)
        # --> att_vec : bsz x n_atoms x (h_heads * d_head)

        return att_vec

class FeedForward(nn.Module):
    def __init__(self, d_model, d_ff):
        super(FeedForward, self).__init__()
        self.l1 = nn.Linear(d_model, d_ff)
        self.l2 = nn.Linear(d_ff, d_model)
        self.ln = nn.LayerNorm(d_model)

    def forward(self, x):
        return self.ln(x + self.l2(F.relu(self.l1(x))))

class AttentionBlock(nn.Module):
    def __init__(self, n_heads, d_model, d_ff, idx, res_p=0.0, att_p=0.0, bias=True, scale=True):
        super(AttentionBlock, self).__init__()
        self.self_att = MultiHeadAttention(n_heads, d_model, idx, res_p, att_p, bias, scale)
        self.ff = FeedForward(d_model, d_ff)
    
    def forward(self, x, pdist, angle, adj, mask=None):
        return self.ff(self.self_att(x, pdist, angle, adj, mask))

class Transformer(nn.Module):
    def __init__(self, n_layers, n_heads, d_model, d_ff, res_p=0.0, att_p=0.0):
        super(Transformer, self).__init__()
        self.layers = nn.ModuleList([
            AttentionBlock(n_heads, d_model, d_ff, idx=i, res_p=res_p, att_p=att_p) for i in range(n_layers)
        ])

    def forward(self, x, pdist, angle, adj, mask):
        for layer in self.layers:
            x = layer(x, pdist, angle, adj, mask=mask)
        return x

class AtomTransformer(nn.Module):
    def __init__(self, n_layers, n_heads, d_model, d_ff, res_p=0.0, att_p=0.0):
        super(AtomTransformer, self).__init__()
        self.d_model = d_model
        self.transformer = Transformer(n_layers, n_heads, d_model, d_ff, res_p, att_p)

        # atom_embedding_size = d_model - 17
        hyb_embedding_size = 64
        donac_embedding_size = 16
        spin_embedding_size = 16
        atom_embedding_size = d_model - 64 - 16 - 16 - 1 - 1 - 1 - 1 - 2
        assert atom_embedding_size > 0, f'Atom embedding size should be > 0. Now {atom_embedding_size}'
        # 1 for aromaticity
        # 1 for formal charge
        # 1 for totalnumH
        # 1 for total valence
        # 2 for molecule features (NPR1, NPR2)
        self.atom_embedding = nn.Embedding(num_unique_atoms + 1 + 1, atom_embedding_size) # +1 for dummy node, +1 for [PAD]
        self.hyb_embedding = nn.Linear(7, hyb_embedding_size, bias=False)
        self.donac_embedding = nn.Linear(2, donac_embedding_size, bias=False)
        self.spin_embedding = nn.Linear(2, spin_embedding_size, bias=False)

        n_head_feature = d_model * 3
        self.gap_head = nn.Sequential(
            nn.Linear(n_head_feature, 16),
            nn.ReLU(),
            nn.Linear(16, 1),
        )

        self.s1_head = nn.Sequential(
            nn.Linear(n_head_feature, 16),
            nn.ReLU(),
            nn.Linear(16, 1),
        )

        self.t1_head = nn.Sequential(
            nn.Linear(n_head_feature, 16),
            nn.ReLU(),
            nn.Linear(16, 1),
        )
    
    # def forward(self, atom_idx, hyb, donac, spin, feat, fp, pdist, sym, angle, adj, mask, out_mask, n_atoms):
    def forward(self, atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms):
        atom = self.atom_embedding(atom_idx + 1)
        hyb = self.hyb_embedding(hyb)
        donac = self.donac_embedding(donac)
        spin = self.spin_embedding(spin)

        x = torch.cat([feat, atom, hyb, donac, spin], dim=-1)
        
        # x : bsz x max_n_atoms x model
        x = self.transformer(x, pdist, angle, adj, mask)

        # Take mean and max only for where atom exists.
        _out_mask = out_mask.unsqueeze(2).expand(x.size(0), x.size(1), x.size(2))
        mean_masked_x = x *_out_mask
        max_masked_x = x - (~_out_mask).float() * 10000
        min_masked_x = x + (~_out_mask).float() * 10000

        x_mean = mean_masked_x.sum(axis=1) / n_atoms.view(x.size(0), 1)
        x_max = max_masked_x.max(axis=1).values
        x_min = min_masked_x.min(axis=1).values

        x = torch.cat([x_mean, x_max, x_min], dim=-1)

        return self.gap_head(x), self.s1_head(x), self.t1_head(x)

class AtomTransformerPretrain(nn.Module):
    def __init__(self, n_layers, n_heads, d_model, d_ff, res_p=0.0, att_p=0.0):
        super(AtomTransformerPretrain, self).__init__()
        self.d_model = d_model
        self.transformer = Transformer(n_layers, n_heads, d_model, d_ff, res_p, att_p)

        feat_size = 6
        hyb_embedding_size = 64
        donac_embedding_size = 16
        spin_embedding_size = 16
        atom_embedding_size = d_model - feat_size - hyb_embedding_size - donac_embedding_size - spin_embedding_size

        assert atom_embedding_size > 0, f'Atom embedding size should be > 0. Now {atom_embedding_size}'
        # 1 for aromaticity
        # 1 for formal charge
        # 1 for totalnumH
        # 1 for total valence
        # 2 for molecule features (NPR1, NPR2)
        self.atom_embedding = nn.Embedding(num_unique_atoms + 1 + 1, atom_embedding_size) # +1 for dummy node, +1 for [PAD]
        self.hyb_embedding = nn.Linear(7, hyb_embedding_size, bias=False)
        self.donac_embedding = nn.Linear(2, donac_embedding_size, bias=False)
        self.spin_embedding = nn.Linear(2, spin_embedding_size, bias=False)

        n_head_feature = d_model * 3

        self.homo_head = nn.Sequential(
            nn.Linear(n_head_feature, 16),
            nn.ReLU(),
            nn.Linear(16, 1),
        )

        self.lumo_head = nn.Sequential(
            nn.Linear(n_head_feature, 16),
            nn.ReLU(),
            nn.Linear(16, 1),
        )
    
    def forward(self, atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms):
        atom = self.atom_embedding(atom_idx + 1)
        hyb = self.hyb_embedding(hyb)
        donac = self.donac_embedding(donac)
        spin = self.spin_embedding(spin)

        x = torch.cat([feat, atom, hyb, donac, spin], dim=-1)
        
        # x : bsz x max_n_atoms x d_model
        x = self.transformer(x, pdist, angle, adj, mask)

        # Take mean and max only for where atom exists.
        _out_mask = out_mask.unsqueeze(2).expand(x.size(0), x.size(1), x.size(2))
        mean_masked_x = x *_out_mask
        max_masked_x = x - (~_out_mask).float() * 10000
        min_masked_x = x + (~_out_mask).float() * 10000

        x_mean = mean_masked_x.sum(axis=1) / n_atoms.view(x.size(0), 1)
        x_max = max_masked_x.max(axis=1).values
        x_min = min_masked_x.min(axis=1).values

        x = torch.cat([x_mean, x_max, x_min], dim=-1)
        return self.homo_head(x), self.lumo_head(x)

if __name__ == '__main__':
    bsz, n_atoms, d_model = 16, max_n_atoms, 128
    
    mask = torch.randint(0, 1, [bsz, n_atoms, n_atoms]).bool().unsqueeze(1) # bsz x 1 x n_atoms x n_atoms
    out_mask = torch.randint(0, 1, [bsz, n_atoms])
    num_atoms = torch.randint(1, 10, [bsz])
    pdist = torch.randn([bsz, n_atoms, n_atoms])
    adj = torch.randn([bsz, n_atoms, n_atoms])
    atom_idx = torch.randint(-1, num_unique_atoms - 1, [bsz, n_atoms])
    feat = torch.randn([bsz, n_atoms, 8])

    xyz = torch.randn([bsz, n_atoms, 3])

    att = AtomTransformer(n_layers=6, n_heads=8, d_model=128, d_ff=1024)
    out = att(xyz, atom_idx, feat, pdist, adj, mask, out_mask, num_atoms)
    print(out)

    print('Model size')
    print(sum(param.numel() for param in att.parameters() if param.requires_grad))



from setuptools import setup, find_packages

setup(
  name = 'sac2021',
  packages = find_packages(exclude=[]),
  include_package_data = True,
  version = '0.0.3',
  license='MIT',
  description = '5th solution for Samsung AI Challenge for Scientific Discovery (2021)',
  author = 'Dohoon Lee',
  author_email = 'dohlee.bioinfo@gmail.com',
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/dohlee/samsung-ai-challenge-solution',
  keywords = [
    'artificial intelligence',
    'cheminformatics',
    'molecular property prediction',
    'quantum chemistry',
  ],
  install_requires=[
    'einops>=0.3',
    'rdkit',
    'numpy',
    'torch>=1.6',
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9',
  ],
)



import wandb

import argparse
import random
import os
import torch
import torch.nn as nn
import numpy as np
import pandas as pd

from torch.utils.data import DataLoader
from sac2021.sac2021 import AtomTransformer
from sac2021.data import SACData
from sklearn.model_selection import KFold

import sac2021.const as const
import sac2021.loss as loss

torch.autograd.set_detect_anomaly(True)

def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--meta', help='Path to metadata csv file.', required=True)
    parser.add_argument('--data', help='Path to directory containing sdf files.', required=True)
    parser.add_argument('--ckpt', required=True)
    parser.add_argument('--output', '-o', required=True)
    parser.add_argument('--model-id', required=True)
    parser.add_argument('--fold', type=int, required=True)
    parser.add_argument('--loss', required=True)
    parser.add_argument('--epochs', type=int, default=200)
    parser.add_argument('--bsz', type=int, default=64)
    parser.add_argument('--n-layers', type=int, default=24)
    parser.add_argument('--d-model', type=int, default=256)
    parser.add_argument('--n-heads', type=int, default=16)
    parser.add_argument('--d-ff', type=int, default=2048)
    parser.add_argument('--lr', type=float, default=3e-5)
    parser.add_argument('--seed', type=int, default=42)
    parser.add_argument('--use-wandb', action='store_true', default=False)
    parser.add_argument('--debug', action='store_true', default=False)

    return parser.parse_args()

args = parse_arguments()

random.seed(args.seed)
np.random.seed(args.seed)
os.environ["PYTHONHASHSEED"] = str(args.seed)
torch.manual_seed(args.seed)
torch.cuda.manual_seed(args.seed)  # type: ignore
# torch.backends.cudnn.deterministic = True  # type: ignore
# torch.backends.cudnn.benchmark = True  # type: ignore

if args.debug:
    os.environ['WANDB_MODE'] = 'offline'

meta = pd.read_csv(args.meta)
meta = meta[~meta.uid.isin(const.ignored_uids)].reset_index(drop=True)

losses = {
    'mse': nn.MSELoss,
    'mae': nn.L1Loss,
    'logcosh': loss.LogCoshLoss,
}
criterion = losses[args.loss]()
mae = nn.L1Loss()

# W&B configuration.
wandb.init(project='sac-solution', entity='dohlee', tags=['finetune'])
config = wandb.config
config.update(args)
config.update({'pretrained': True})
config.update({'pretrained_model': args.ckpt})
config.update({'data_version': SACData.version})

# Train-val split (39:1)
cv = KFold(n_splits=40, shuffle=True, random_state=args.seed)
for fold, (train_idx, val_idx) in enumerate(cv.split(meta), 1):
    if fold == args.fold:
        break

val_idx_set = set(val_idx)
train_idx = [idx for idx in range(len(meta)) if idx not in val_idx_set]

print('train', len(train_idx))
print('val', len(val_idx))

net = AtomTransformer(
    config.n_layers,
    config.n_heads,
    config.d_model,
    config.d_ff
)

# Load checkpoint.
ckpt = torch.load(args.ckpt)
net.load_state_dict(ckpt['net'], strict=False)

# Freeze parameters. Will be unfreezed later.
for param in net.atom_embedding.parameters():
    param.requires_grad = False
for param in net.hyb_embedding.parameters():
    param.requires_grad = False
for param in net.donac_embedding.parameters():
    param.requires_grad = False
for param in net.spin_embedding.parameters():
    param.requires_grad = False
for param in net.transformer.parameters():
    param.requires_grad = False

net.cuda()

wandb.watch(net, log='all')
optimizer = torch.optim.AdamW(net.parameters(), lr=args.lr)
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
    optimizer, mode='min', factor=0.5, 
    patience=15, threshold=0.005, threshold_mode='abs'
)

train_dataset = SACData(
    meta=args.meta,
    data=args.data,
    idx=train_idx,
)
val_dataset = SACData(
    meta=args.meta,
    data=args.data,
    idx=val_idx
)
train_loader = DataLoader(
    train_dataset, num_workers=16, batch_size=config.bsz, shuffle=True, pin_memory=True,
)
val_loader = DataLoader(
    val_dataset, num_workers=16, batch_size=config.bsz, shuffle=False, pin_memory=True, drop_last=False,
)
n_batch = len(train_loader)

optimizer.zero_grad()
optimizer.step()

best_val_mae, val_mae = 10000, 10000
for epoch in range(1, 201):
    scheduler.step(val_mae)

    # Unfreeze weights after 9 epochs
    if epoch == 10:
        for param in net.atom_embedding.parameters():
            param.requires_grad = True
        for param in net.hyb_embedding.parameters():
            param.requires_grad = True
        for param in net.donac_embedding.parameters():
            param.requires_grad = True
        for param in net.spin_embedding.parameters():
            param.requires_grad = True
        for param in net.transformer.parameters():
            param.requires_grad = True

    # losses
    running_gap_loss = 0.0
    running_s1_loss = 0.0
    running_t1_loss = 0.0

    # metrics (MAE)
    running_gap_mae = 0.0
    running_gap_mae_secondary = 0.0
    running_gap_mae_tertiary = 0.0

    # Train
    net.train()
    for batch, data in enumerate(train_loader, 1):
        for k, v in data.items():
            data[k] = v.cuda()

        optimizer.zero_grad()

        atom_idx = data['atom_idx']
        hyb = data['hyb']
        donac = data['donac']
        spin = data['spin']
        feat = data['feat']
        pdist = data['pdist']
        angle = data['angle']
        adj = data['adj']
        mask = data['mask']
        out_mask = data['out_mask']
        n_atoms = data['n_atoms']
        gap_target = data['target'].float()
        s1_target = data['s1'].float()
        t1_target = data['t1'].float()

        gap_out, s1_out, t1_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)
        _bsz = gap_out.size(0)
        gap_loss = criterion(gap_out, gap_target.view(_bsz, 1))
        s1_loss = criterion(s1_out, s1_target.view(_bsz, 1))
        t1_loss = criterion(t1_out, t1_target.view(_bsz, 1))

        loss = gap_loss + 0.05 * (s1_loss + t1_loss)

        running_gap_loss += gap_loss.detach().item()
        running_s1_loss += s1_loss.detach().item()
        running_t1_loss += t1_loss.detach().item()

        _mae = mae(gap_out, gap_target.view(_bsz, 1))
        running_gap_mae += _mae.detach().item()

        _mae = mae(s1_out - t1_out, gap_target.view(_bsz, 1))
        running_gap_mae_secondary += _mae.detach().item()

        _mae = mae(0.5 * gap_out + 0.5 * (s1_out - t1_out) , gap_target.view(_bsz, 1))
        running_gap_mae_tertiary += _mae.detach().item()

        loss.backward()
        nn.utils.clip_grad_norm_(net.parameters(), 1)
        optimizer.step()

        if batch % 10 == 0:
            batch_gap_loss = running_gap_loss / 10.
            batch_s1_loss = running_s1_loss / 10.
            batch_t1_loss = running_t1_loss / 10.
            batch_gap_mae = running_gap_mae / 10.
            batch_gap_mae_secondary = running_gap_mae_secondary / 10.
            batch_gap_mae_tertiary = running_gap_mae_tertiary / 10.

            print(f'E{epoch:<3}B{batch:<3} l={batch_gap_loss:.4f} l2={batch_s1_loss:.4f} l3={batch_t1_loss:.4f} mae={batch_gap_mae:.4f} mae2={batch_gap_mae_secondary:.4f} mae3={batch_gap_mae_tertiary:.4f} lmae={np.log(batch_gap_mae):.4f}')

            running_gap_loss = 0.0
            running_s1_loss = 0.0
            running_t1_loss = 0.0
            running_gap_mae = 0.0
            running_gap_mae_secondary = 0.0
            running_gap_mae_tertiary = 0.0
            
            log_dict = {
                'train/loss': batch_gap_loss,
                'train/s1_loss': batch_s1_loss,
                'train/t1_loss': batch_t1_loss,
                'train/mae': batch_gap_mae, 
                'train/mae_secondary': batch_gap_mae_secondary,
                'train/mae_tertiary': batch_gap_mae_tertiary,
                'train/lr': get_lr(optimizer)
            }
            wandb.log(log_dict)

    # Validate
    net.eval()
    val_gap_outs, val_s1_outs, val_t1_outs = [], [], []
    val_gap_targets, val_s1_targets, val_t1_targets = [], [], []
    with torch.no_grad():
        for batch, data in enumerate(val_loader, 1):
            for k, v in data.items():
                data[k] = v.cuda()

            atom_idx = data['atom_idx']
            hyb = data['hyb']
            donac = data['donac']
            spin = data['spin']
            feat = data['feat']
            pdist = data['pdist']
            angle = data['angle']
            adj = data['adj']
            mask = data['mask']
            out_mask = data['out_mask']
            n_atoms = data['n_atoms']
            target = data['target'].float()
            s1 = data['s1'].float()
            t1 = data['t1'].float()

            gap_out, s1_out, t1_out = net(atom_idx, hyb, donac, spin, feat, pdist, angle, adj, mask, out_mask, n_atoms)

            _bsz = gap_out.size(0)
            val_gap_outs.append(gap_out.cpu())
            val_s1_outs.append(s1_out.cpu())    
            val_t1_outs.append(t1_out.cpu())
            val_gap_targets.append(target.cpu().view(_bsz, -1))
            val_s1_targets.append(s1.cpu().view(_bsz, -1))
            val_t1_targets.append(t1.cpu().view(_bsz, -1))

    val_gap_outs, val_s1_outs, val_t1_outs = map(lambda x: torch.cat(x, dim=0), (val_gap_outs, val_s1_outs, val_t1_outs))
    val_gap_targets, val_s1_targets, val_t1_targets = map(lambda x: torch.cat(x, dim=0), (val_gap_targets, val_s1_targets, val_t1_targets))

    val_gap_loss = criterion(val_gap_outs, val_gap_targets)
    val_s1_loss = criterion(val_s1_outs, val_s1_targets)
    val_t1_loss = criterion(val_t1_outs, val_t1_targets)

    val_mae = mae(val_gap_outs, val_gap_targets)
    val_mae_secondary = mae(val_s1_outs - val_t1_outs, val_gap_targets)
    val_mae_tertiary = mae(0.5 * val_gap_outs + 0.5 * (val_s1_outs - val_t1_outs), val_gap_targets)

    print(f'***Validation E{epoch:<3} loss={val_gap_loss:.4f} s1_loss={val_s1_loss:.4f} t1_loss={val_t1_loss:.4f} mae={val_mae:.4f} mae2={val_mae_secondary:.4f} mae3={val_mae_tertiary:.4f} lmae={np.log(val_mae):.4f}')
    log_dict = {
        'val/loss': val_gap_loss,
        'val/mae': val_mae,
        'val/mae_secondary': val_mae_secondary,
        'val/mae_tertiary': val_mae_tertiary,
        'epoch': epoch,
    }
    wandb.log(log_dict)

    if val_mae < best_val_mae:
        best_val_mae = val_mae
        wandb.summary['best_val_mae'] = best_val_mae

        ckpt = {
            'net': net.state_dict(),
            'optimizer': optimizer.state_dict(),
            'epoch': epoch,
            'val_mae': best_val_mae,
        }
        torch.save(ckpt, args.output)
